""" Some classes used for MultiROI scan processing. """
import numpy as np


class ROI:
    """ Holds ROI info and computes an xy plane (scanfield) at a given z.

    ScanImage defines a ROI as the interpolation between a set of scanfields. See their
    docs for details.
    """
    def __init__(self, roi_info):
        """ Read the scanfields that define this ROI and other required info.

        Args:
            roi_info: A dictionary containing the definition of the roi extracted from the
                tiff header.
        """
        self.roi_info = roi_info
        self._scanfields = None

    @property
    def scanfields(self):
        if self._scanfields is None:
            self._scanfields = self._create_scanfields()
        return self._scanfields

    @property
    def is_discrete_plane_mode_on(self):
        return bool(self.roi_info['discretePlaneMode'])

    def _create_scanfields(self):
        """ Create all the scanfields that form this ROI. """
        # Get scanfield configuration info
        scanfield_infos = self.roi_info['scanfields']
        if not isinstance(scanfield_infos, list):
            scanfield_infos = [scanfield_infos] # make list if single scanfield

        # Get scanfield depths
        scanfield_depths = self.roi_info['zs']
        if not isinstance(scanfield_depths, list):
            scanfield_depths = [scanfield_depths]

        scanfields = []
        for scanfield_info, scanfield_depth in zip(scanfield_infos, scanfield_depths):
            # if scanfield_info['enable']: # this is always 1 even if ROI is disabled
            # Get scanfield info
            width, height = scanfield_info['pixelResolutionXY']
            xcenter, ycenter = scanfield_info['centerXY']
            size_in_x, size_in_y = scanfield_info['sizeXY']

            # Create scanfield
            new_scanfield = Scanfield(height=height, width=width, depth=scanfield_depth,
                                      y=ycenter, x=xcenter, height_in_degrees=size_in_y,
                                      width_in_degrees=size_in_x)
            scanfields.append(new_scanfield)

        # Sort them by depth (to ease interpolation)
        scanfields = sorted(scanfields, key=lambda scanfield: scanfield.depth)

        return scanfields

    def get_field_at(self, scanning_depth):
        """ Interpolates between the ROI scanfields to generate the 2-d field at the
        desired depth.

        Args:
            scanning_depth: An integer. Depth at which we want to obtain the field.

        Returns:
            field. A Field object, the desired field at scanning_depth

        Warning:
            Does not work for rotated ROIs.
            If there were more than one scanfield at the same depth, it will only consider
                the one defined last.
        """
        field = None

        if self.is_discrete_plane_mode_on: # only check at each scanfield depth
            for scanfield in self.scanfields:
                if scanning_depth == scanfield.depth:
                    field = scanfield.as_field()
        else:
            if len(self.scanfields) == 1: # single scanfield extending from -inf to inf
                field = self.scanfields[0].as_field()
                field.depth = scanning_depth

            else: # interpolate between scanfields
                scanfield_depths = [sf.depth for sf in self.scanfields]
                valid_range = range(min(scanfield_depths), max(scanfield_depths) + 1)
                if scanning_depth in valid_range:
                    field = Field()

                    scanfield_heights = [sf.height for sf in self.scanfields]
                    field.height = np.interp(scanning_depth, scanfield_depths,
                                             scanfield_heights)
                    field.height = int(round(field.height / 2)) * 2 # round to the closest even

                    scanfield_widths = [sf.width for sf in self.scanfields]
                    field.width = np.interp(scanning_depth, scanfield_depths,
                                            scanfield_widths)
                    field.width = int(round(field.width / 2)) * 2 # round to the closest even

                    field.depth = scanning_depth

                    scanfield_ys = [sf.y for sf in self.scanfields]
                    field.y = np.interp(scanning_depth, scanfield_depths, scanfield_ys)

                    scanfield_xs = [sf.x for sf in self.scanfields]
                    field.x = np.interp(scanning_depth, scanfield_depths, scanfield_xs)

                    scanfield_heights = [sf.height_in_degrees for sf in self.scanfields]
                    field.height_in_degrees = np.interp(scanning_depth, scanfield_depths,
                                                        scanfield_heights)

                    scanfield_widths = [sf.width_in_degrees for sf in self.scanfields]
                    field.width_in_degrees = np.interp(scanning_depth, scanfield_depths,
                                                       scanfield_widths)

        return field


class Scanfield:
    """ Small container for scanfield information. Used to define ROIs.

    Attributes:
        height: height of the field in pixels.
        width: width of the field in pixels.
        depth: depth at which this field was recorded (in microns relative to absolute z).
        y, x: Coordinates of the center of the field in the scan (in scan angle degrees).
        height_in_degrees: height of the field in degrees of the scan angle.
        width_in_degrees: width of the field in degrees of the scan angle.
    """
    def __init__(self, height=None, width=None, depth=None, y=None, x=None,
                 height_in_degrees=None, width_in_degrees=None):
        self.height = height
        self.width = width
        self.depth = depth
        self.y = y
        self.x = x
        self.height_in_degrees = height_in_degrees
        self.width_in_degrees = width_in_degrees

    def as_field(self):
        return Field(height=self.height, width=self.width, depth=self.depth, y=self.y,
                     x=self.x, height_in_degrees=self.height_in_degrees,
                     width_in_degrees=self.width_in_degrees)


class Field(Scanfield):
    """ Two-dimensional scanning plane. An extension of scanfield with some functionality.

    Attributes:
        height: height of the field in pixels.
        width: width of the field in pixels.
        depth: depth at which this field was recorded (in microns relative to absolute z).
        y, x: Coordinates of the center of the field in the scan (in scan angle degrees).
        height_in_degrees: height of the field in degrees of the scan angle.
        width_in_degrees: width of the field in degrees of the scan angle.
        yslices: list of slices. How to slice the page in the y axis to get this field.
        xslices: list of slices. How to slice the page in the x axis to get this field.
            For now, all fields have the same width so all xslices are slice(None).
        output_yslices: list of slices. Where to paste this field in the output field.
        output_xslices: list of slices. Where to paste this field in the output field.
        slice_id: index of the slice in the scan to which this field belongs.
        roi_ids: list of ROI indices to which each subfield belongs (one if single field).
        offsets: list of masks with time offsets per pixel (seconds, one if single field).

    Example:
        output_field[output_yslice, output_xslice] = page[yslice, xslice]

    When a field is formed by joining two or more subfields (via join_contiguous), the
    slice lists hold two or more slices representing where each subfield will be taken
    from the page and inserted in the (joint) output field. Attributes height, width, x,
    y, height_in_degrees and width_in_degrees are adjusted accordingly. For non-contiguous
    fields, each slice list has a single slice.

     Note:
        Slices in xslices, yslices, output_xslices and output_yslices hold two promises:
            step = 1 (fields are contiguous)
            stop = start + height|width
        In theory, we only need x_start and y_start but slices simplify operations.

    """
    def __init__(self, height=None, width=None, depth=None, y=None, x=None,
                 height_in_degrees=None, width_in_degrees=None, yslices=None,
                 xslices=None, output_yslices=None, output_xslices=None, slice_id=None,
                 roi_ids=None, offsets=None):
        self.height = height
        self.width = width
        self.depth = depth
        self.y = y
        self.x = x
        self.height_in_degrees = height_in_degrees
        self.width_in_degrees = width_in_degrees
        self.yslices = yslices
        self.xslices = xslices
        self.output_yslices = output_yslices
        self.output_xslices = output_xslices
        self.slice_id = slice_id
        self.roi_ids = roi_ids
        self.offsets = offsets

    @property
    def has_contiguous_subfields(self):
        """ Whether field is formed by many contiguous subfields. """
        return len(self.xslices) > 1

    @property
    def roi_mask(self):
        """ Mask of the size of the field. Each pixel shows the ROI from where it comes."""
        mask = np.full([self.height, self.width], -1, dtype=np.int8)
        for roi_id, output_yslice, output_xslice in zip(self.roi_ids, self.output_yslices,
                                                        self.output_xslices):
            mask[output_yslice, output_xslice] = roi_id
        return mask

    @property
    def offset_mask(self):
        """ Mask of the size of the field. Each pixel shows its time offset in seconds."""
        mask = np.full([self.height, self.width], -1, dtype=np.float32)
        for offsets, output_yslice, output_xslice in zip(self.offsets, self.output_yslices,
                                                        self.output_xslices):
            mask[output_yslice, output_xslice] = offsets
        return mask

    def _type_of_contiguity(self, field2):
        """ Compute how field 2 is contiguous to this one.

        Args:
            field2: A second field object.

        Returns:
            An integer {NONCONTIGUOUS = 0, ABOVE = 1, BELOW = 2, LEFT = 3, RIGHT = 4}.
               Whether field 2 is above, below, to the left or to the right of this field.
        """
        position = Position.NONCONTIGUOUS
        if np.isclose(self.width_in_degrees, field2.width_in_degrees):
            expected_distance = self.height_in_degrees / 2 + field2.height_in_degrees / 2
            if np.isclose(self.y, field2.y + expected_distance):
                position = Position.ABOVE
            if np.isclose(field2.y, self.y + expected_distance):
                position = Position.BELOW
        if np.isclose(self.height_in_degrees, field2.height_in_degrees):
            expected_distance = self.width_in_degrees / 2 + field2.width_in_degrees / 2
            if np.isclose(self.x, field2.x + expected_distance):
                position = Position.LEFT
            if np.isclose(field2.x, self.x + expected_distance):
                position = Position.RIGHT

        return position

    def is_contiguous_to(self, field2):
        """ Whether this field is contiguous to field2."""
        return not (self._type_of_contiguity(field2) == Position.NONCONTIGUOUS)

    def join_with(self, field2):
        """ Update attributes of this field to incorporate field2. Field2 is NOT changed.

        Args:
            field2: A second field object.
        """
        contiguity = self._type_of_contiguity(field2)
        if contiguity in [Position.ABOVE, Position.BELOW]:  # contiguous in y axis
            # Compute some specific attributes
            if contiguity == Position.ABOVE:  # field2 is above/atop self
                new_y = field2.y + (self.height_in_degrees / 2)
                output_yslices1 = [slice(s.start + field2.height, s.stop + field2.height)
                                   for s in self.output_yslices]
                output_yslices2 = field2.output_yslices
            else:  # field2 is below self
                new_y = self.y + (field2.height_in_degrees / 2)
                output_yslices1 = self.output_yslices
                output_yslices2 = [slice(s.start + self.height, s.stop + self.height) for
                                   s in field2.output_yslices]

            # Set new attributes
            self.y = new_y
            self.height += field2.height
            self.height_in_degrees += field2.height_in_degrees
            self.output_yslices = output_yslices1 + output_yslices2
            self.output_xslices = self.output_xslices + field2.output_xslices

        if contiguity in [Position.LEFT, Position.RIGHT]:  # contiguous in x axis
            # Compute some specific attributes
            if contiguity == Position.LEFT:  # field2 is to the left of self
                new_x = field2.x + (self.width_in_degrees / 2)
                output_xslices1 = [slice(s.start + field2.width, s.stop + field2.width)
                                   for s in self.output_xslices]
                output_xslices2 = field2.output_xslices
            else:  # field2 is to the right of self
                new_x = self.x + (field2.width_in_degrees / 2)
                output_xslices1 = self.output_xslices
                output_xslices2 = [slice(s.start + self.width, s.stop + self.width) for s
                                   in field2.output_xslices]

            # Set new attributes
            self.x = new_x
            self.width += field2.width
            self.width_in_degrees += field2.width_in_degrees
            self.output_yslices = self.output_yslices + field2.output_yslices
            self.output_xslices = output_xslices1 + output_xslices2

        # yslices and xslices just get appended regardless of the type of contiguity
        self.yslices = self.yslices + field2.yslices
        self.xslices = self.xslices + field2.xslices

        # Append roi ids and offsets
        self.roi_ids = self.roi_ids + field2.roi_ids
        self.offsets = self.offsets + field2.offsets


class Position:
    NONCONTIGUOUS = 0
    ABOVE = 1
    BELOW = 2
    LEFT = 3
    RIGHT = 4


import os

import numpy as np
import tifffile
from pathlib import Path


def make_json_serializable(obj):
    """Convert metadata to JSON serializable format."""
    if isinstance(obj, dict):
        return {k: make_json_serializable(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [make_json_serializable(v) for v in obj]
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, (np.integer, np.floating)):
        return obj.item()
    else:
        return obj


def get_metadata(file: os.PathLike | str):
    """
    Extract metadata from a TIFF file. This can be a raw ScanImage TIFF or one
    processed via [lbm_caiman_python.save_as()](#save_as).

    Parameters
    ----------
    file: os.PathLike
        Path to the TIFF file.

    Returns
    -------
    dict
        Metadata extracted from the TIFF file.

    Raises
    ------
    ValueError
        If no metadata is found in the TIFF file. This can occur when the file is not a ScanImage TIFF.
    """
    if not file:
        return None

    tiff_file = tifffile.TiffFile(file)
    if (
            hasattr(tiff_file, 'shaped_metadata')
            and tiff_file.shaped_metadata is not None
            and isinstance(tiff_file.shaped_metadata, (list, tuple))
            and tiff_file.shaped_metadata
            and tiff_file.shaped_metadata[0] not in ([], (), None)
    ):
        if 'image' in tiff_file.shaped_metadata[0]:
            return tiff_file.shaped_metadata[0]['image']

    if hasattr(tiff_file, 'scanimage_metadata'):
        meta = tiff_file.scanimage_metadata
        if meta is None:
            return None

        si = meta.get('FrameData', {})
        if not si:
            print(f"No FrameData found in {file}.")
            return None

        series = tiff_file.series[0]
        pages = tiff_file.pages

        # Extract ROI and imaging metadata
        roi_group = meta["RoiGroups"]["imagingRoiGroup"]["rois"]

        num_rois = len(roi_group)
        num_planes = len(si["SI.hChannels.channelSave"])
        scanfields = roi_group[0]["scanfields"]  # assuming single ROI scanfield configuration

        # ROI metadata
        size_xy = scanfields["sizeXY"]
        num_pixel_xy = scanfields["pixelResolutionXY"]

        # TIFF header-derived metadata
        sample_format = pages[0].dtype.name
        objective_resolution = si["SI.objectiveResolution"]
        frame_rate = si["SI.hRoiManager.scanFrameRate"]

        # Field-of-view calculations
        # TODO: We may want an FOV measure that takes into account contiguous ROIs
        # As of now, this is for a single ROI
        fov_x = round(objective_resolution * size_xy[0])
        fov_y = round(objective_resolution * size_xy[1])
        fov_xy = (fov_x, int(fov_y / num_rois))

        # Pixel resolution (dxy) calculation
        pixel_resolution = (fov_x / num_pixel_xy[0], fov_y / num_pixel_xy[1])

        return {
            "num_planes": num_planes,
            "num_frames": int(len(pages) / num_planes),
            "fov": fov_xy,  # in microns
            "num_rois": num_rois,
            "frame_rate": frame_rate,
            "pixel_resolution": np.round(pixel_resolution, 2),
            "ndim": series.ndim,
            "dtype": 'uint16',
            "size": series.size,
            "raw_height": pages[0].shape[0],
            "raw_width": pages[0].shape[1],
            "tiff_pages": len(pages),
            "roi_width_px": num_pixel_xy[0],
            "roi_height_px": num_pixel_xy[1],
            "sample_format": sample_format,
            "objective_resolution": objective_resolution,
        }
    else:
        raise ValueError(f"No metadata found in {file}.")


def get_files_ext(base_dir, extension, max_depth) -> list:
    """
    Recursively searches for files with a specific extension up to a given depth and stores their paths in a pickle file.

    Parameters
    ----------
    base_dir : str or Path
        The base directory to start searching.
    extension : str
        The file extension to look for (e.g., '.txt').
    max_depth : int
        The maximum depth of subdirectories to search.

    Returns
    -------
    list
        A list of full file paths matching the given extension.
    """
    base_path = Path(base_dir).expanduser().resolve()
    if not base_path.exists():
        raise FileNotFoundError(f"Directory '{base_path}' does not exist.")
    if not base_path.is_dir():
        raise NotADirectoryError(f"'{base_path}' is not a directory.")

    return [
        str(file)
        for file in base_path.rglob(f'*{extension}')
        if len(file.relative_to(base_path).parts) <= max_depth + 1
    ]



"""
quick_metrics.py
"""
import numpy as np
from skimage.restoration import estimate_sigma


def estimate_noise(image):
    """
    Estimate the noise in a 2D array.
    
    Parameters
    ----------
    image : 2D numpy array
        The array for which to estimate the noise.
        
    Returns
    -------
    sigma : float or list
        Estimated noise standard deviation(s).  If `multichannel` is True and
        `average_sigmas` is False, a separate noise estimate for each channel
        is returned.  Otherwise, the average of the individual channel
        estimates is returned.

    Notes
    -----
    This function assumes the noise follows a Gaussian distribution. The
    estimation algorithm is based on the median absolute deviation of the
    wavelet detail coefficients as described in section 4.2 of [1]_.

    References
    ----------
    .. [1] D. L. Donoho and I. M. Johnstone. "Ideal spatial adaptation
       by wavelet shrinkage." Biometrika 81.3 (1994): 425-455.
       :DOI:`10.1093/biomet/81.3.425`

    Examples
    --------
    >>> rng = np.random.default_rng()
    >>> img = img + sigma * rng.standard_normal(img.shape)
    >>> sigma_hat = estimate_noise(img)
    """
    return estimate_sigma(image, multichannel=True, average_sigmas=True)



import argparse
import functools
import os
import time
import warnings
from pathlib import Path
import numpy as np
import zarr
from scanreader import read_scan
from scanreader.utils import listify_index
from .lcp_io import get_metadata, make_json_serializable

import tifffile
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

ARRAY_METADATA = ["dtype", "shape", "nbytes", "size"]

CHUNKS = {0: 'auto', 1: -1, 2: -1}

# https://brainglobe.info/documentation/brainglobe-atlasapi/adding-a-new-atlas.html
BRAINGLOBE_STRUCTURE_TEMPLATE = {
    "acronym": "VIS",  # shortened name of the region
    "id": 3,  # region id
    "name": "visual cortex",  # full region name
    "structure_id_path": [1, 2, 3],  # path to the structure in the structures hierarchy, up to current id
    "rgb_triplet": [255, 255, 255],
    # default color for visualizing the region, feel free to leave white or randomize it
}

# suppress warnings
warnings.filterwarnings("ignore")

print = functools.partial(print, flush=True)


def process_slice_str(slice_str):
    if not isinstance(slice_str, str):
        raise ValueError(f"Expected a string argument, received: {slice_str}")
    if slice_str.isdigit():
        return int(slice_str)
    else:
        parts = slice_str.split(":")
    return slice(*[int(p) if p else None for p in parts])


def process_slice_objects(slice_str):
    return tuple(map(process_slice_str, slice_str.split(",")))


def print_params(params, indent=5):
    for k, v in params.items():
        # if value is a dictionary, recursively call the function
        if isinstance(v, dict):
            print(" " * indent + f"{k}:")
            print_params(v, indent + 4)
        else:
            print(" " * indent + f"{k}: {v}")


def return_scan_offset(image_in, nvals: int = 8):
    """
    Compute the scan offset correction between interleaved lines or columns in an image.

    This function calculates the scan offset correction by analyzing the cross-correlation
    between interleaved lines or columns of the input image. The cross-correlation peak
    determines the amount of offset between the lines or columns, which is then used to
    correct for any misalignment in the imaging process.

    Parameters
    ----------
    image_in : ndarray | ndarray-like
        Input image or volume. It can be 2D, 3D, or 4D.

    .. note::

        Dimensions: [height, width], [time, height, width], or [time, plane, height, width].
        The input array must be castable to numpy. e.g. np.shape, np.ravel.

    nvals : int
        Number of pixel-wise shifts to include in the search for best correlation.

    Returns
    -------
    int
        The computed correction value, based on the peak of the cross-correlation.

    Examples
    --------
    >>> img = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
    >>> return_scan_offset(img, 1)

    Notes
    -----
    This function assumes that the input image contains interleaved lines or columns that
    need to be analyzed for misalignment. The cross-correlation method is sensitive to
    the similarity in pattern between the interleaved lines or columns. Hence, a strong
    and clear peak in the cross-correlation result indicates a good alignment, and the
    corresponding lag value indicates the amount of misalignment.
    """
    from scipy import signal

    image_in = image_in.squeeze()

    if len(image_in.shape) == 3:
        image_in = np.mean(image_in, axis=0)
    elif len(image_in.shape) == 4:
        image_in = np.mean(np.mean(image_in, axis=0), axis=0)

    n = nvals

    in_pre = image_in[::2, :]
    in_post = image_in[1::2, :]

    min_len = min(in_pre.shape[0], in_post.shape[0])
    in_pre = in_pre[:min_len, :]
    in_post = in_post[:min_len, :]

    buffers = np.zeros((in_pre.shape[0], n))

    in_pre = np.hstack((buffers, in_pre, buffers))
    in_post = np.hstack((buffers, in_post, buffers))

    in_pre = in_pre.T.ravel(order="F")
    in_post = in_post.T.ravel(order="F")

    # Zero-center and clip negative values to zero
    # Iv1 = Iv1 - np.mean(Iv1)
    in_pre[in_pre < 0] = 0

    in_post = in_post - np.mean(in_post)
    in_post[in_post < 0] = 0

    in_pre = in_pre[:, np.newaxis]
    in_post = in_post[:, np.newaxis]

    r_full = signal.correlate(in_pre[:, 0], in_post[:, 0], mode="full", method="auto")
    unbiased_scale = len(in_pre) - np.abs(np.arange(-len(in_pre) + 1, len(in_pre)))
    r = r_full / unbiased_scale

    mid_point = len(r) // 2
    lower_bound = mid_point - n
    upper_bound = mid_point + n + 1
    r = r[lower_bound:upper_bound]
    lags = np.arange(-n, n + 1)

    # Step 3: Find the correction value
    correction_index = np.argmax(r)
    return lags[correction_index]


def fix_scan_phase(
        data_in: np.ndarray,
        offset: int,
):
    """
    Corrects the scan phase of the data based on a given offset along a specified dimension.

    Parameters:
    -----------
    dataIn : ndarray
        The input data of shape (sy, sx, sc, sz).
    offset : int
        The amount of offset to correct for.

    Returns:
    --------
    ndarray
        The data with corrected scan phase, of shape (sy, sx, sc, sz).
    """
    dims = data_in.shape
    ndim = len(dims)
    if ndim == 2:
        sy, sx = data_in.shape
        data_out = np.zeros_like(data_in)

        if offset > 0:
            # Shift even df left and odd df right by 'offset'
            data_out[0::2, :sx - offset] = data_in[0::2, offset:]
            data_out[1::2, offset:] = data_in[1::2, :sx - offset]
        elif offset < 0:
            offset = abs(offset)
            # Shift even df right and odd df left by 'offset'
            data_out[0::2, offset:] = data_in[0::2, :sx - offset]
            data_out[1::2, :sx - offset] = data_in[1::2, offset:]
        else:
            print("Phase = 0, no correction applied.")
            return data_in

        return data_out
    if ndim == 4:
        st, sc, sy, sx = data_in.shape
        if offset != 0:
            data_out = np.zeros((st, sc, sy, sx + abs(offset)))
        else:
            print("Phase = 0, no correction applied.")
            return data_in

        if offset > 0:
            data_out[:, :, 0::2, :sx] = data_in[:, :, 0::2, :]
            data_out[:, :, 1::2, offset: offset + sx] = data_in[:, :, 1::2, :]
            data_out = data_out[:, :, :, : sx + offset]
        elif offset < 0:
            offset = abs(offset)
            data_out[:, :, 0::2, offset: offset + sx] = data_in[:, :, 0::2, :]
            data_out[:, :, 1::2, :sx] = data_in[:, :, 1::2, :]
            data_out = data_out[:, :, :, offset:]

        return data_out

    if ndim == 3:
        st, sy, sx = data_in.shape
        if offset != 0:
            # Create output array with appropriate shape adjustment
            data_out = np.zeros((st, sy, sx + abs(offset)))
        else:
            print("Phase = 0, no correction applied.")
            return data_in

        if offset > 0:
            # For positive offset
            data_out[:, 0::2, :sx] = data_in[:, 0::2, :]
            data_out[:, 1::2, offset: offset + sx] = data_in[:, 1::2, :]
            # Trim output by excluding columns that contain only zeros
            data_out = data_out[:, :, : sx + offset]
        elif offset < 0:
            # For negative offset
            offset = abs(offset)
            data_out[:, 0::2, offset: offset + sx] = data_in[:, 0::2, :]
            data_out[:, 1::2, :sx] = data_in[:, 1::2, :]
            # Trim output by excluding the first 'offset' columns
            data_out = data_out[:, :, offset:]

        return data_out

    raise NotImplementedError()


def save_as(
        scan,
        savedir: os.PathLike,
        planes=None,
        frames=None,
        metadata=None,
        overwrite=True,
        ext='.tiff',
):
    """
    Save scan data to the specified directory in the desired format.

    Parameters
    ----------
    scan : scanreader.ScanMultiROI
        An object representing scan data. Must have attributes such as `num_channels`,
        `num_frames`, `fields`, and `rois`, and support indexing for retrieving frame data.
    savedir : os.PathLike
        Path to the directory where the data will be saved.
    planes : int, list, or tuple, optional
        Plane indices to save. If `None`, all planes are saved. Default is `None`.
    frames : list or tuple, optional
        Frame indices to save. If `None`, all frames are saved. Default is `None`.
    metadata : dict, optional
        Additional metadata to update the scan object's metadata. Default is `None`.
    overwrite : bool, optional
        Whether to overwrite existing files. Default is `True`.
    ext : str, optional
        File extension for the saved data. Supported options are `'.tiff'` and `'.zarr'`.
        Default is `'.tiff'`.

    Raises
    ------
    ValueError
        If an unsupported file extension is provided.

    Notes
    -----
    This function creates the specified directory if it does not already exist.
    Data is saved per channel, organized by planes.
    """

    savedir = Path(savedir)
    if planes is None:
        planes = list(range(scan.num_channels))
    elif not isinstance(planes, (list, tuple)):
        planes = [planes]
    if frames is None:
        frames = list(range(scan.num_frames))
    elif not isinstance(planes, (list, tuple)):
        frames = [frames]
    if not metadata:
        metadata = {'si': scan.tiff_files[0].scanimage_metadata,
                    'image': make_json_serializable(get_metadata(scan.tiff_files[0].filehandle.path))}

    if not savedir.exists():
        logger.debug(f"Creating directory: {savedir}")
        savedir.mkdir(parents=True)
    _save_data(scan, savedir, planes, frames, overwrite, ext, metadata)


def _save_data(scan, path, planes, frames, overwrite, file_extension, metadata):
    path.mkdir(parents=True, exist_ok=True)
    print(f'Planes: {planes}')

    file_writer = _get_file_writer(file_extension, overwrite, metadata)
    if len(scan.fields) > 1:
        for idx, field in enumerate(scan.fields):
            for chan in planes:
                if 'tif' in file_extension:
                    arr = scan[idx, :, :, chan, frames]  # [y,x,T]
                    logger.debug('arr shape:', arr.shape)
                    file_writer(path, f'plane_{chan + 1}_roi_{idx + 1}', arr.T)
    else:
        for chan in planes:
            if 'tif' in file_extension:
                arr = scan[:, :, :, chan, frames]  # [y,x,T]
                logger.debug('arr shape:', arr.shape)
                file_writer(path, f'plane_{chan + 1}', arr.T)


def _get_file_writer(ext, overwrite, metadata=None):
    if ext in ['.tif', '.tiff']:
        return functools.partial(_write_tiff, overwrite=overwrite, metadata=metadata)
    elif ext == '.zarr':
        return functools.partial(_write_zarr, overwrite=overwrite, metadata=metadata)
    else:
        raise ValueError(f'Unsupported file extension: {ext}')


def _write_tiff(path, name, data, overwrite=True, metadata=None):
    filename = Path(path / f'{name}.tiff')
    if filename.exists() and not overwrite:
        logger.warning(
            f'File already exists: {filename}. To overwrite, set overwrite=True (--overwrite in command line)')
        return
    logger.info(f"Writing {filename}")
    t_write = time.time()
    data = np.transpose(data.squeeze(), (0, 2, 1))
    tifffile.imwrite(filename, data, metadata=metadata)
    t_write_end = time.time() - t_write
    logger.info(f"Data written in {t_write_end:.2f} seconds.")


def _write_zarr(path, name, data, metadata=None, overwrite=True):
    store = zarr.DirectoryStore(path)
    root = zarr.group(store, overwrite=overwrite)
    ds = root.create_dataset(name=name, data=data.squeeze(), overwrite=True)
    if metadata:
        ds.attrs['metadata'] = metadata


def main():
    parser = argparse.ArgumentParser(description="CLI for processing ScanImage tiff files.")
    parser.add_argument("path",
                        type=str,
                        nargs='?',  # Change this to make 'path' optional
                        default=None,
                        help="Path to the file or directory to process.")
    parser.add_argument("--frames",
                        type=str,
                        default=":",  # all frames
                        help="Frames to read (0 based). Use slice notation like NumPy arrays ("
                             "e.g., :50 gives frames 0 to 50, 5:15:2 gives frames 5 to 15 in steps of 2)."
                        )
    parser.add_argument("--planes",
                        type=str,
                        default=":",  # all planes
                        help="Planes to read (0 based). Use slice notation like NumPy arrays (e.g., 1:5 gives planes "
                             "2 to 6")
    parser.add_argument("--trimx",
                        type=int,
                        nargs=2,
                        default=(0, 0),
                        help="Number of x-pixels to trim from each ROI. Tuple or list (e.g., 4 4 for left and right "
                             "edges).")
    parser.add_argument("--trimy", type=int, nargs=2, default=(0, 0),
                        help="Number of y-pixels to trim from each ROI. Tuple or list (e.g., 4 4 for top and bottom "
                             "edges).")
    # Boolean Flags
    parser.add_argument("--metadata", action="store_true",
                        help="Print a dictionary of scanimage metadata for files at the given path.")
    parser.add_argument("--roi",
                        action='store_true',
                        help="Save each ROI in its own folder, organized like 'zarr/roi_1/plane_1/, without this "
                             "arguemnet it would save like 'zarr/plane_1/roi_1'."
                        )

    parser.add_argument("--save", type=str, nargs='?', help="Path to save data to. If not provided, the path will be "
                                                            "printed.")
    parser.add_argument("--overwrite", action='store_true', help="Overwrite existing files if saving data..")
    parser.add_argument("--tiff", action='store_false', help="Flag to save as .tiff. Default is True")
    parser.add_argument("--zarr", action='store_true', help="Flag to save as .zarr. Default is False")
    parser.add_argument("--assemble", action='store_true', help="Flag to assemble the each ROI into a single image.")
    parser.add_argument("--debug", action='store_true', help="Output verbose debug information.")
    parser.add_argument("--delete_first_frame", action='store_false', help="Flag to delete the first frame of the "
                                                                           "scan when saving.")
    # Commands
    args = parser.parse_args()

    # If no arguments are provided, print help and exit
    if len(vars(args)) == 0 or not args.path:
        parser.print_help()
        return

    if args.debug:
        logger.setLevel(logging.DEBUG)
        logger.debug("Debug mode enabled.")

    path = Path(args.path).expanduser()
    if path.is_dir():
        files = [str(x) for x in Path(args.path).expanduser().glob('*.tif*')]
    elif path.is_file():
        files = [str(path)]
    else:
        raise FileNotFoundError(f"File or directory not found: {args.path}")

    if len(files) < 1:
        raise ValueError(
            f"Input path given is a non-tiff file: {args.path}.\n"
            f"scanreader is currently limited to scanimage .tiff files."
        )
    else:
        print(f'Found {len(files)} file(s) in {args.path}')

    if args.metadata:
        t_metadata = time.time()
        metadata = get_metadata(files[0])
        t_metadata_end = time.time() - t_metadata
        print(f"Metadata read in {t_metadata_end:.2f} seconds.")
        print(f"Metadata for {files[0]}:")
        # filter out the verbose scanimage frame/roi metadata
        print_params({k: v for k, v in metadata.items() if k not in ['si', 'roi_info']})

    if args.assemble:
        join_contiguous = True
    else:
        join_contiguous = False

    if args.save:
        savepath = Path(args.save).expanduser()
        logger.info(f"Saving data to {savepath}.")

        t_scan_init = time.time()
        scan = read_scan(files, join_contiguous=join_contiguous, )
        t_scan_init_end = time.time() - t_scan_init
        logger.info(f"--- Scan initialized in {t_scan_init_end:.2f} seconds.")

        frames = listify_index(process_slice_str(args.frames), scan.num_frames)
        zplanes = listify_index(process_slice_str(args.planes), scan.num_channels)

        if args.delete_first_frame:
            frames = frames[1:]
            logger.debug(f"Deleting first frame. New frames: {frames}")

        logger.debug(f"Frames: {len(frames)}")
        logger.debug(f"Z-Planes: {len(zplanes)}")

        if args.zarr:
            ext = '.zarr'
            logger.debug("Saving as .zarr.")
        elif args.tiff:
            ext = '.tiff'
            logger.debug("Saving as .tiff.")
        else:
            raise NotImplementedError("Only .zarr and .tif are supported file formats.")

        t_save = time.time()
        save_as(
            scan,
            savepath,
            frames=frames,
            planes=zplanes,
            overwrite=args.overwrite,
            ext=ext,
        )
        t_save_end = time.time() - t_save
        logger.info(f"--- Processing complete in {t_save_end:.2f} seconds. --")
        return scan
    else:
        print(args.path)


if __name__ == '__main__':
    main()




"""
ScanImage scans. Each version handles things a little differently. Scan objects are
usually instantiated by a call to scanreader.read_scan().

Hierarchy:
BaseScan
    ScanLegacy
    BaseScan5
        Scan5Point1
        Scan5Point2
            Scan5Point3
                Scan5Point4
                Scan5Point5
                Scan5Point6
                Scan5Point7
                Scan2016b
                Scan2017a
                Scan2017b
                Scan2018a
                Scan2018b
                Scan2019a
                Scan2019b
                Scan2020
    ScanMultiRoi
"""
from tifffile import TiffFile
from tifffile.tifffile import matlabstr2py
import numpy as np
import re
import itertools
from . import utils
from .multiroi import ROI
from .exceptions import FieldDimensionMismatch

class BaseScan():
    """ Properties and methods shared among all scan versions.

    Scan objects are a collection of recording fields: rectangular planes at a given x, y,
    z position in the scan recorded in a number of channels during a preset amount of
    time. All fields have the same number of channels and number of frames.
    Scan objects are:
        indexable: scan[field, y, x, channel, frame] works as long as the fields' spatial
            dimensions (y, x) match.
        iterable: 'for field in scan:' iterates over all fields (4-d array) in the scan.

    Examples:
        scan.version                ScanImage version of the scan.
        scan[:, :, :3, :, :1000]    5-d numpy array with the first 1000 frames of the
            first 3 fields (if x, y dimensions match).
        for field in scan:          generates 4-d numpy arrays ([y, x, channels, frames]).

    Note:
        We use the word 'frames' as in video frames, i.e., number of timesteps the scan
        was recorded; ScanImage uses frames to refer to slices/scanning depths in the
        scan.
    """
    """
    Interface rules:
        If it (method or property) is shared among all subclasses it should be here,
            either implemented or as an abstract method. Even if private (e.g. _page_height)
        If one subclass needs to overwrite it, then erase it here and implement them in
            the subclasses (this applies for now that I only have two subclasses). If code
            needs to be shared add it as a private method here.
        If it is not in every subclass, it should not be here.
    """
    def __init__(self):
        self.filenames = None
        self.dtype = None
        self._tiff_files = None
        self.header = ''

    @property
    def tiff_files(self):
        if self._tiff_files is None:
            self._tiff_files = [TiffFile(filename) for filename in self.filenames]
        return self._tiff_files

    @tiff_files.deleter
    def tiff_files(self):
        if self._tiff_files is not None:
            for tiff_file in self._tiff_files:
                tiff_file.close()
            self._tiff_files = None

    @property
    def version(self):
        match = re.search(r"SI.?\.VERSION_MAJOR = '?(?P<version>[^\s']*)'?", self.header)
        version = match.group('version') if match else None
        return version

    @property
    def is_slow_stack(self):
        """ True if fastZ is disabled. All frames for one slice are recorded first before
        moving to the next slice."""
        match = re.search(r'hFastZ\.enable = (?P<is_slow>.*)', self.header)
        is_slow_stack = (match.group('is_slow') in ['false', '0']) if match else None
        return is_slow_stack

    @property
    def is_multiROI(self):
        """Only True if mroiEnable exists (2016b and up) and is set to True."""
        match = re.search(r'hRoiManager\.mroiEnable = (?P<is_multiROI>.)', self.header)
        is_multiROI = (match.group('is_multiROI') == '1') if match else False
        return is_multiROI

    @property
    def num_channels(self):
        match = re.search(r'hChannels\.channelSave = (?P<channels>.*)', self.header)
        if match:
            channels = matlabstr2py(match.group('channels'))
            num_channels = len(channels) if isinstance(channels, list) else 1
        else:
            num_channels = None
        return num_channels

    @property
    def requested_scanning_depths(self):
        match = re.search(r'hStackManager\.zs = (?P<zs>.*)', self.header)
        if match:
            zs = matlabstr2py(match.group('zs'))
            scanning_depths = zs if isinstance(zs, list) else [zs]
        else:
            scanning_depths = None
        return scanning_depths

    @property
    def num_scanning_depths(self):
        if self.is_slow_stack:
            """ Number of scanning depths actually recorded in this stack."""
            num_scanning_depths = self._num_pages / (self.num_channels * self.num_frames)
            num_scanning_depths = int(num_scanning_depths) # discard last slice if incomplete
        else:
            num_scanning_depths = len(self.requested_scanning_depths)
        return num_scanning_depths

    @property
    def scanning_depths(self):
        return self.requested_scanning_depths[:self.num_scanning_depths]

    @property
    def num_requested_frames(self):
        if self.is_slow_stack:
             match = re.search(r'hStackManager\.framesPerSlice = (?P<num_frames>.*)',
                              self.header)
        else:
            match = re.search(r'hFastZ\.numVolumes = (?P<num_frames>.*)', self.header)
        num_requested_frames = int(1e9 if match.group('num_frames')=='Inf' else
                                   float(match.group('num_frames'))) if match else None
        return num_requested_frames

    @property
    def num_frames(self):
        """ Each tiff page is an image at a given channel, scanning depth combination."""
        if self.is_slow_stack:
            num_frames = min(self.num_requested_frames / self._num_averaged_frames,
                             self._num_pages / self.num_channels) # finished in the first slice
        else:
            num_frames = self._num_pages / (self.num_channels * self.num_scanning_depths)
        num_frames = int(num_frames) # discard last frame if incomplete
        return num_frames

    @property
    def is_bidirectional(self):
        match = re.search(r'hScan2D\.bidirectional = (?P<is_bidirectional>.*)', self.header)
        is_bidirectional = (match.group('is_bidirectional') == 'true') if match else False
        return is_bidirectional

    @property
    def scanner_frequency(self):
        match = re.search(r'hScan2D\.scannerFrequency = (?P<scanner_freq>.*)', self.header)
        scanner_frequency = float(match.group('scanner_freq')) if match else None
        return scanner_frequency

    @property
    def seconds_per_line(self):
        if np.isnan(self.scanner_frequency):
            match = re.search(r'hRoiManager\.linePeriod = (?P<secs_per_line>.*)', self.header)
            seconds_per_line = float(match.group('secs_per_line')) if match else None
        else:
            scanner_period = 1 / self.scanner_frequency # secs for mirror to return to initial position
            seconds_per_line = scanner_period / 2 if self.is_bidirectional else scanner_period
        return seconds_per_line

    @property
    def _num_pages(self):
        num_pages = sum([len(tiff_file.pages) for tiff_file in self.tiff_files])
        return num_pages

    @property
    def _page_height(self):
        return self.tiff_files[0].pages[0].imagelength

    @property
    def _page_width(self):
        return self.tiff_files[0].pages[0].imagewidth

    @property
    def _num_averaged_frames(self):
        """ Number of requested frames are averaged to form one saved frame. """
        match = re.search(r'hScan2D\.logAverageFactor = (?P<num_avg_frames>.*)', self.header)
        num_averaged_frames = int(float(match.group('num_avg_frames'))) if match else None
        return num_averaged_frames

    @property
    def num_fields(self):
        raise NotImplementedError('Subclasses of BaseScan must implement this property')

    @property
    def field_depths(self):
        raise NotImplementedError('Subclasses of BaseScan must implement this property')

    # Properties from here on are not strictly necessary
    @property
    def fps(self):
        match = re.search(r'hRoiManager\.scanVolumeRate = (?P<fps>.*)',self.header)
        fps = float(match.group('fps')) if match else None
        return fps

    @property
    def spatial_fill_fraction(self):
        match = re.search(r'hScan2D\.fillFractionSpatial = (?P<spatial_ff>.*)', self.header)
        spatial_fill_fraction = float(match.group('spatial_ff')) if match else None
        return spatial_fill_fraction

    @property
    def temporal_fill_fraction(self):
        match = re.search(r'hScan2D\.fillFractionTemporal = (?P<temporal_ff>.*)', self.header)
        temporal_fill_fraction = float(match.group('temporal_ff')) if match else None
        return temporal_fill_fraction

    @property
    def scanner_type(self):
        match = re.search(r"hScan2D\.scannerType = '(?P<scanner_type>.*)'", self.header)
        scanner_type = match.group('scanner_type') if match else None
        return scanner_type

    @property
    def motor_position_at_zero(self):
        """ Motor position (x, y and z in microns) corresponding to the scan's (0, 0, 0)
        point. For non-multiroi scans, (x=0, y=0) marks the center of the FOV."""
        match = re.search(r'hMotors\.motorPosition = (?P<motor_position>.*)', self.header)
        motor_position = matlabstr2py(match.group('motor_position'))[:3] if match else None
        return motor_position

    @property
    def initial_secondary_z(self):
        """ Initial position in z (microns) of the secondary motor (if any)."""
        match = re.search(r'hMotors\.motorPosition = (?P<motor_position>.*)', self.header)
        if match:
            motor_position = matlabstr2py(match.group('motor_position'))
            secondary_z = motor_position[3] if len(motor_position) > 3 else None
        else:
            secondary_z = None
        return secondary_z

    @property
    def _initial_frame_number(self):
        match = re.search(r'\sframeNumbers = (?P<frame_number>.*)', self.header)
        initial_frame_number = int(match.group('frame_number')) if match else None
        return initial_frame_number

    @property
    def _num_fly_back_lines(self):
        """ Lines/mirror cycles that it takes to move from one depth to the next."""
        match = re.search(r'hScan2D\.flybackTimePerFrame = (?P<fly_back_seconds>.*)',
                          self.header)
        if match:
            fly_back_seconds = float(match.group('fly_back_seconds'))
            num_fly_back_lines = self._seconds_to_lines(fly_back_seconds)
        else:
            num_fly_back_lines = None
        return num_fly_back_lines

    @property
    def _num_lines_between_fields(self):
        """ Lines/mirror cycles scanned from the start of one field to the start of the
        next. """
        if self.is_slow_stack:
            num_lines_between_fields = ((self._page_height + self._num_fly_back_lines) *
                                        (self.num_frames * self._num_averaged_frames))
        else:
            num_lines_between_fields = self._page_height + self._num_fly_back_lines
        return num_lines_between_fields

    @property
    def is_slow_stack_with_fastZ(self):
        raise NotImplementedError('Subclasses of BaseScan must implement this property')

    @property
    def field_offsets(self):
        raise NotImplementedError('Subclasses of BaseScan must implement this property')

    def read_data(self, filenames, dtype):
        """ Set self.header, self.filenames and self.dtype. Data is read lazily when needed.

        Args:
            filenames: List of strings. Tiff filenames.
            dtype: Data type of the output array.
        """
        self.filenames = filenames # set filenames
        self.dtype=dtype # set dtype of read data
        self.header = '{}\n{}'.format(self.tiff_files[0].pages[0].description,
                                      self.tiff_files[0].pages[0].software) # set header (ScanImage metadata)

    def __array__(self):
        return self[:]

    def __str__(self):
        msg = '{}\n{}\n{}'.format(type(self), '*' * 80, self.header, '*' * 80)
        return msg

    def __len__(self):
        return 0 if self.num_fields is None else self.num_fields

    def __getitem__(self, key):
        """ Index scans by field, y, x, channels, frames. Supports integer, slice and
        array/tuple/list of integers as indices."""
        raise NotImplementedError('Subclasses of BaseScan must implement this method')

    def __iter__(self):
        class ScanIterator:
            """ Iterator for Scan objects."""
            def __init__(self, scan):
                self.scan = scan
                self.next_field = 0

            def __next__(self):
                if self.next_field < self.scan.num_fields:
                    field = self.scan[self.next_field]
                    self.next_field += 1
                else:
                    raise StopIteration
                return field

        return ScanIterator(self)

    def _read_pages(self, slice_list, channel_list, frame_list, yslice=slice(None),
                    xslice=slice(None)):
        """ Reads the tiff pages with the content of each slice, channel, frame
        combination and slices them in the y, x dimension.

        Each tiff page holds a single depth/channel/frame combination. For slow stacks,
        channels change first, timeframes change second and slices/depths change last.
        Example:
            For two channels, three slices, two frames.
                Page:       0   1   2   3   4   5   6   7   8   9   10  11
                Channel:    0   1   0   1   0   1   0   1   0   1   0   1
                Frame:      0   0   1   1   2   2   0   0   1   1   2   2
                Slice:      0   0   0   0   0   0   1   1   1   1   1   1

        For scans, channels change first, slices/depths change second and timeframes
        change last.
        Example:
            For two channels, three slices, two frames.
                Page:       0   1   2   3   4   5   6   7   8   9   10  11
                Channel:    0   1   0   1   0   1   0   1   0   1   0   1
                Slice:      0   0   1   1   2   2   0   0   1   1   2   2
                Frame:      0   0   0   0   0   0   1   1   1   1   1   1

        Args:
            slice_list: List of integers. Slices to read.
            channel_list: List of integers. Channels to read.
            frame_list: List of integers. Frames to read
            yslice: Slice object. How to slice the pages in the y axis.
            xslice: Slice object. How to slice the pages in the x axis.

        Returns:
            A 5-D array (num_slices, output_height, output_width, num_channels, num_frames).
                Required pages reshaped to have slice, channel and frame as different
                dimensions. Channel, slice and frame order received in the input lists are
                respected; for instance, if slice_list = [1, 0, 2, 0], then the first
                dimension will have four slices: [1, 0, 2, 0].

        Note:
            We use slices in y, x for memory efficiency, If lists were passed another copy
            of the pages will be needed coming up to 3x the amount of data we actually
            want to read (the output array, the read pages and the list-sliced pages).
            Slices limit this to 2x (output array and read pages which are sliced in place).
        """
        # Compute pages to load from tiff files
        if self.is_slow_stack:
            frame_step = self.num_channels
            slice_step = self.num_channels * self.num_frames
        else:
            slice_step = self.num_channels
            frame_step = self.num_channels * self.num_scanning_depths
        pages_to_read = []
        for frame in frame_list:
            for slice_ in slice_list:
                for channel in channel_list:
                    new_page = frame * frame_step + slice_ * slice_step + channel
                    pages_to_read.append(new_page)

        # Compute output dimensions
        out_height = len(utils.listify_index(yslice, self._page_height))
        out_width = len(utils.listify_index(xslice, self._page_width))

        # Read pages
        pages = np.empty([len(pages_to_read), out_height, out_width], dtype=self.dtype)
        start_page = 0
        for tiff_file in self.tiff_files:

            # Get indices in this tiff file and in output array
            final_page_in_file = start_page + len(tiff_file.pages)
            is_page_in_file = lambda page: page in range(start_page, final_page_in_file)
            pages_in_file = filter(is_page_in_file, pages_to_read)
            file_indices = [page - start_page for page in pages_in_file]
            global_indices = [is_page_in_file(page) for page in pages_to_read]

            # Read from this tiff file (if needed)
            if len(file_indices) > 0:
                # this line looks a bit ugly but is memory efficient. Do not separate
                pages[global_indices] = tiff_file.asarray(key=file_indices)[..., yslice, xslice]
            start_page += len(tiff_file.pages)

        # Reshape the pages into (slices, y, x, channels, frames)
        new_shape = [len(frame_list), len(slice_list), len(channel_list), out_height, out_width]
        pages = pages.reshape(new_shape).transpose([1, 3, 4, 2, 0])

        return pages

    def _seconds_to_lines(self, seconds):
        """ Compute how many lines would be scanned in the given amount of seconds."""
        num_lines = int(np.ceil(seconds / self.seconds_per_line))
        if self.is_bidirectional:
            # scanning starts at one end of the image so num_lines needs to be even
            num_lines += (num_lines % 2)

        return num_lines

    def _compute_offsets(self, field_height, start_line):
        """ Computes the time offsets at which a given field was recorded.

        Computes the time delay at which each pixel was recorded using the start of the
        scan as zero. It first creates an image with the number of lines scanned until
        that point and then uses self.seconds_per_line to  transform it into seconds.

        :param int field_height: Height of the field.
        :param int start_line: Line at which this field starts.

        :returns: A field_height x page_width mask with time offsets in seconds.
        """
        # Compute offsets within a line (negligible if seconds_per_line is small)
        max_angle = (np.pi / 2) * self.temporal_fill_fraction
        line_angles = np.linspace(-max_angle, max_angle, self._page_width + 2)[1:-1]
        line_offsets = (np.sin(line_angles) + 1) / 2

        # Compute offsets for entire field
        field_offsets = np.expand_dims(np.arange(0, field_height), -1) + line_offsets
        if self.is_bidirectional: # odd lines scanned from left to right
            field_offsets[1::2] = field_offsets[1::2] - line_offsets + (1 - line_offsets)

        # Transform offsets from line counts to seconds
        field_offsets = (field_offsets + start_line) * self.seconds_per_line

        return field_offsets


class ScanLegacy(BaseScan):
    """ Scan versions 4 and below. Not implemented."""

    def __init__(self):
        raise NotImplementedError('Legacy scans not supported')


class BaseScan5(BaseScan):
    """ ScanImage 5 scans: one field per scanning depth and all fields have the same
    height and width."""

    @property
    def num_fields(self):
        return self.num_scanning_depths # one field per scanning depth

    @property
    def field_depths(self):
        return self.scanning_depths

    @property
    def image_height(self):
        return self._page_height

    @property
    def image_width(self):
        return self._page_width

    @property
    def shape(self):
        return (self.num_fields, self.image_height, self.image_width, self.num_channels,
                self.num_frames)

    @property
    def zoom(self):
        match = re.search(r'hRoiManager\.scanZoomFactor = (?P<zoom>.*)', self.header)
        zoom = float(match.group('zoom')) if match else None
        return zoom

    @property
    def is_slow_stack_with_fastZ(self):
        match = re.search(r'hMotors\.motorSecondMotorZEnable = (?P<uses_fastZ>.*)',
                          self.header)
        uses_fastZ = (match.group('uses_fastZ') in ['true', '1']) if match else None
        return self.is_slow_stack and uses_fastZ

    @property
    def field_offsets(self):
        """ Seconds elapsed between start of frame scanning and each pixel."""
        next_line = 0
        field_offsets = []
        for i in range(self.num_fields):
            field_offsets.append(self._compute_offsets(self.image_height, next_line))
            next_line += self._num_lines_between_fields
        return field_offsets

    @property
    def _y_angle_scale_factor(self):
        """ Scan angles in y are scaled by this factor, shrinking the angle range."""
        match = re.search(r'hRoiManager\.scanAngleMultiplierSlow = (?P<angle_scaler>.*)',
                          self.header)
        y_angle_scaler = float(match.group('angle_scaler')) if match else None
        return y_angle_scaler

    @property
    def _x_angle_scale_factor(self):
        """ Scan angles in x are scaled by this factor, shrinking the angle range."""
        match = re.search(r'hRoiManager\.scanAngleMultiplierFast = (?P<angle_scaler>.*)',
                          self.header)
        x_angle_scaler = float(match.group('angle_scaler')) if match else None
        return x_angle_scaler

    def __getitem__(self, key):
        """ In non-multiROI, all fields have the same x, y dimensions. """
        # Fill key to size 5 (raises IndexError if more than 5)
        full_key = utils.fill_key(key, num_dimensions=5)

        # Check index types are valid
        for i, index in enumerate(full_key):
            utils.check_index_type(i, index)

        # Check each dimension is in bounds
        max_dimensions = self.shape
        for i, (index, dim_size) in enumerate(zip(full_key, max_dimensions)):
            utils.check_index_is_in_bounds(i, index, dim_size)

        # Get fields, channels and frames as lists
        field_list = utils.listify_index(full_key[0], self.num_fields)
        y_list = utils.listify_index(full_key[1], self.image_height)
        x_list = utils.listify_index(full_key[2], self.image_width)
        channel_list = utils.listify_index(full_key[3], self.num_channels)
        frame_list = utils.listify_index(full_key[4], self.num_frames)

        # Edge case when slice index gives 0 elements or index is empty list, e.g., scan[10:0], scan[[]]
        if [] in [field_list, y_list, x_list, channel_list, frame_list,]:
            return np.empty(0)

        # Read the required pages
        pages = self._read_pages(field_list, channel_list, frame_list)

        # Index in y, x using the original key (usually slices) for memory efficiency.
        if isinstance(full_key[1], list) and isinstance(full_key[2], list):
            # Our behaviour for lists is to take the submatrix defined by those indices.
            ys = [[y] for y in y_list] # ys as nested lists does the trick
            item = pages[:, ys, x_list, :, :]
        else:
            item = pages[:, full_key[1], full_key[2], :, :]
            item = item.reshape(len(field_list), len(y_list), len(x_list), len(channel_list),
                                len(frame_list)) # put back any dropped dimension

        # If original index was an integer, delete that axis (as in numpy indexing)
        squeeze_dims = [i for i, index in enumerate(full_key) if np.issubdtype(type(index),
                                                                               np.signedinteger)]
        item = np.squeeze(item, axis=tuple(squeeze_dims))

        return item


class Scan5Point1(BaseScan5):
    """ ScanImage 5.1. Basic."""
    pass


class Scan5Point2(BaseScan5):
    """ ScanImage 5.2. Addition of FOV measures in microns."""

    @property
    def image_height_in_microns(self):
        match = re.search(r'hRoiManager\.imagingFovUm = (?P<fov_corners>.*)', self.header)
        if match:
            fov_corners = matlabstr2py(match.group('fov_corners'))
            image_height_in_microns = fov_corners[2][1] - fov_corners[1][1]  # y1-y0
        else:
            image_height_in_microns = None
        return image_height_in_microns

    @property
    def image_width_in_microns(self):
        match = re.search(r'hRoiManager\.imagingFovUm = (?P<fov_corners>.*)', self.header)
        if match:
            fov_corners = matlabstr2py(match.group('fov_corners'))
            image_width_in_microns = fov_corners[1][0] - fov_corners[0][0] # x1-x0
        else:
            image_width_in_microns = None
        return image_width_in_microns


class NewerScan():
    """ Shared features among all newer scans. """
    @property
    def is_slow_stack_with_fastZ(self):
        match = re.search(r'hStackManager\.slowStackWithFastZ = (?P<slow_with_fastZ>.*)',
                          self.header)
        slow_with_fastZ = (match.group('slow_with_fastZ') in ['true', '1']) if match else None
        return slow_with_fastZ


class Scan5Point3(NewerScan, Scan5Point2): # NewerScan first to shadow Scan5Point2's properties
    """ScanImage 5.3"""
    pass

class Scan5Point4(Scan5Point3):
    """ScanImage 5.4"""
    pass

class Scan5Point5(Scan5Point3):
    """ScanImage 5.5"""
    pass

class Scan5Point6(Scan5Point3):
    """ScanImage 5.6"""
    pass

class Scan5Point7(Scan5Point3):
    """ScanImage 5.7"""
    pass


class Scan2016b(Scan5Point3):
    """ ScanImage 2016b"""
    pass


class Scan2017a(Scan5Point3):
    """ ScanImage 2017a"""
    pass


class Scan2017b(Scan5Point3):
    """ ScanImage 2017b"""
    pass


class Scan2018a(Scan5Point3):
    """ ScanImage 2018a"""
    pass


class Scan2018b(Scan5Point3):
    """ ScanImage 2018b"""
    pass


class Scan2019a(Scan5Point3):
    """ ScanImage 2019a"""
    pass


class Scan2019b(Scan5Point3):
    """ ScanImage 2019b"""
    pass

class Scan2020(Scan5Point3):
    """ ScanImage 2020"""
    pass

class Scan2021(Scan5Point3):
    """ ScanImage 2021"""
    pass


class ScanMultiROI(NewerScan, BaseScan):
    """An extension of ScanImage v5 that manages multiROI data (output from mesoscope).

     Attributes:
         join_contiguous: A bool. Whether contiguous fields are joined into one.
         rois: List of ROI objects (defined in multiroi.py)
         fields: List of Field objects (defined in multiroi.py)
     """

    def __init__(self, join_contiguous):
        super().__init__()
        self.join_contiguous = join_contiguous
        self.rois = None
        self.fields = None

    @property
    def num_fields(self):
        return len(self.fields)

    @property
    def num_rois(self):
        return len(self.rois)

    @property
    def field_heights(self):
        return [field.height for field in self.fields]

    @property
    def field_widths(self):
        return [field.width for field in self.fields]

    @property
    def field_depths(self):
        return [field.depth for field in self.fields]

    @property
    def field_slices(self):
        return [field.slice_id for field in self.fields]

    @property
    def field_rois(self):
        return [field.roi_ids for field in self.fields]

    @property
    def field_masks(self):
        return [field.roi_mask for field in self.fields]

    @property
    def field_offsets(self):
        return [field.offset_mask for field in self.fields]

    @property
    def field_heights_in_microns(self):
        field_heights_in_degrees = [field.height_in_degrees for field in self.fields]
        return [self._degrees_to_microns(deg) for deg in field_heights_in_degrees]

    @property
    def field_widths_in_microns(self):
        field_widths_in_degrees = [field.width_in_degrees for field in self.fields]
        return [self._degrees_to_microns(deg) for deg in field_widths_in_degrees]

    @property
    def _num_fly_to_lines(self):
        """ Number of lines recorded in the tiff page while flying to a different field,
        i.e., distance between fields in the tiff page."""
        match = re.search(r'hScan2D\.flytoTimePerScanfield = (?P<fly_to_seconds>.*)',
                          self.header)
        if match:
            fly_to_seconds = float(match.group('fly_to_seconds'))
            num_fly_to_lines = self._seconds_to_lines(fly_to_seconds)
        else:
            num_fly_to_lines = None
        return num_fly_to_lines

    def _degrees_to_microns(self, degrees):
        """ Convert scan angle degrees to microns using the objective resolution."""
        match = re.search(r'objectiveResolution = (?P<deg2um_factor>.*)', self.header)
        microns = (degrees * float(match.group('deg2um_factor'))) if match else None
        return microns

    def read_data(self, filenames, dtype):
        """ Set the header, create rois and fields (joining them if necessary)."""
        super().read_data(filenames, dtype)
        self.rois = self._create_rois()
        self.fields = self._create_fields()
        if self.join_contiguous:
            self._join_contiguous_fields()

    def _create_rois(self):
        """Create scan rois from the configuration file. """
        roi_infos = self.tiff_files[0].scanimage_metadata['RoiGroups']['imagingRoiGroup']['rois']
        roi_infos = roi_infos if isinstance(roi_infos, list) else [roi_infos]
        roi_infos = list(filter(lambda r: isinstance(r['zs'], (int, float, list)),
                                roi_infos)) # discard empty/malformed ROIs

        rois = [ROI(roi_info) for roi_info in roi_infos]
        return rois

    def _create_fields(self):
        """ Go over each slice depth and each roi generating the scanned fields. """
        fields = []
        previous_lines = 0
        for slice_id, scanning_depth in enumerate(self.scanning_depths):
            next_line_in_page = 0 # each slice is one tiff page
            for roi_id, roi in enumerate(self.rois):
                new_field = roi.get_field_at(scanning_depth)

                if new_field is not None:
                    if next_line_in_page + new_field.height > self._page_height:
                        error_msg = ('Overestimated number of fly to lines ({}) at '
                                     'scanning depth {}'.format(self._num_fly_to_lines,
                                                                scanning_depth))
                        raise RuntimeError(error_msg)

                    # Set xslice and yslice (from where in the page to cut it)
                    new_field.yslices = [slice(next_line_in_page, next_line_in_page
                                               + new_field.height)]
                    new_field.xslices = [slice(0, new_field.width)]

                    # Set output xslice and yslice (where to paste it in output)
                    new_field.output_yslices = [slice(0, new_field.height)]
                    new_field.output_xslices = [slice(0, new_field.width)]

                    # Set slice and roi id
                    new_field.slice_id = slice_id
                    new_field.roi_ids = [roi_id]

                    # Set timing offsets
                    offsets = self._compute_offsets(new_field.height, previous_lines +
                                                                      next_line_in_page)
                    new_field.offsets = [offsets]

                    # Compute next starting y
                    next_line_in_page += new_field.height + self._num_fly_to_lines

                    # Add field to fields
                    fields.append(new_field)

            # Accumulate overall number of scanned lines
            previous_lines += self._num_lines_between_fields

        return fields

    def _join_contiguous_fields(self):
        """ In each scanning depth, join fields that are contiguous.

        Fields are considered contiguous if they appear next to each other and have the
        same size in their touching axis. Process is iterative: it tries to join each
        field with the remaining ones (checked in order); at the first union it will break
        and restart the process at the first field. When two fields are joined, it deletes
        the one appearing last and modifies info such as field height, field width and
        slices in the one appearing first.

        Any rectangular area in the scan formed by the union of two or more fields which
        have been joined will be treated as a single field after this operation.
        """
        for scanning_depth in self.scanning_depths:
            two_fields_were_joined = True
            while two_fields_were_joined: # repeat until no fields were joined
                two_fields_were_joined = False

                fields = filter(lambda field: field.depth == scanning_depth, self.fields)
                for field1, field2 in itertools.combinations(fields, 2):

                    if field1.is_contiguous_to(field2):
                        # Change info in field 1 to reflect the union
                        field1.join_with(field2)

                        # Delete field 2 in self.fields
                        self.fields.remove(field2)

                        # Restart join contiguous search (at while)
                        two_fields_were_joined = True
                        break

    def __getitem__(self, key):
        # Fill key to size 5 (raises IndexError if more than 5)
        full_key = utils.fill_key(key, num_dimensions=5)

        # Check index types are valid
        for i, index in enumerate(full_key):
            utils.check_index_type(i, index)

        # Check each dimension is in bounds
        utils.check_index_is_in_bounds(0, full_key[0], self.num_fields)
        for field_id in utils.listify_index(full_key[0], self.num_fields):
            utils.check_index_is_in_bounds(1, full_key[1], self.field_heights[field_id])
            utils.check_index_is_in_bounds(2, full_key[2], self.field_widths[field_id])
        utils.check_index_is_in_bounds(3, full_key[3], self.num_channels)
        utils.check_index_is_in_bounds(4, full_key[4], self.num_frames)

        # Get fields, channels and frames as lists
        field_list = utils.listify_index(full_key[0], self.num_fields)
        y_lists = [utils.listify_index(full_key[1], self.field_heights[field_id]) for
                   field_id in field_list]
        x_lists = [utils.listify_index(full_key[2], self.field_widths[field_id]) for
                   field_id in field_list]
        channel_list = utils.listify_index(full_key[3], self.num_channels)
        frame_list = utils.listify_index(full_key[4], self.num_frames)

        # Edge case when slice index gives 0 elements or index is empty list, e.g., scan[10:0], scan[[]]
        if [] in [field_list, *y_lists, *x_lists, channel_list, frame_list]:
            return np.empty(0)

        # Check output heights and widths match for all fields
        if not all(len(y_list) == len(y_lists[0]) for y_list in y_lists):
            raise FieldDimensionMismatch('Image heights for all fields do not match')
        if not all(len(x_list) == len(x_lists[0]) for x_list in x_lists):
            raise FieldDimensionMismatch('Image widths for all fields do not match')

        # Over each field, read required pages and slice
        item = np.empty([len(field_list), len(y_lists[0]), len(x_lists[0]),
                        len(channel_list), len(frame_list)], dtype=self.dtype)
        for i, (field_id, y_list, x_list) in enumerate(zip(field_list, y_lists, x_lists)):
            field = self.fields[field_id]

            # Over each subfield in field (only one for non-contiguous fields)
            slices = zip(field.yslices, field.xslices, field.output_yslices, field.output_xslices)
            for yslice, xslice, output_yslice, output_xslice in slices:

                # Read the required pages (and slice out the subfield)
                pages = self._read_pages([field.slice_id], channel_list, frame_list,
                                         yslice, xslice)

                # Get x, y indices that need to be accessed in this subfield
                y_range = range(output_yslice.start, output_yslice.stop)
                x_range = range(output_xslice.start, output_xslice.stop)
                ys = [[y - output_yslice.start] for y in y_list if y in y_range]
                xs = [x - output_xslice.start for x in x_list if x in x_range]
                output_ys = [[index] for index, y in enumerate(y_list) if y in y_range]
                output_xs = [index for index, x in enumerate(x_list) if x in x_range]
                # ys as nested lists are needed for numpy to slice them correctly

                # Index pages in y, x
                item[i, output_ys, output_xs] = pages[0, ys, xs]

        # If original index was an integer, delete that axis (as in numpy indexing)
        squeeze_dims = [i for i, index in enumerate(full_key) if np.issubdtype(type(index),
                                                                               np.signedinteger)]
        item = np.squeeze(item, axis=tuple(squeeze_dims))

        return item


class ScanReaderException(Exception):
    """Base ScanReader exception. """
    pass

class ScanImageVersionError(ScanReaderException):
    """ Exception for unsupported ScanImage versions."""
    pass

class PathnameError(ScanReaderException):
    """ Exception for dealing with paths and pathname patterns (wildcards)."""
    pass

class FieldDimensionMismatch(ScanReaderException):
    """ Exception for trying to slice an array with fields of different dimensions."""
    pass


"""
Reader for ScanImage 5 scans (including multiROI).

Example:
    import scanreader
    scan = scanreader.read_scan('my_scan_*.tif')
    for field in scan:
        #process field
"""
from tifffile import TiffFile
from glob import glob
from os import path
import numpy as np
import re
from .exceptions import ScanImageVersionError, PathnameError
from . import scans

_scans = {'5.1': scans.Scan5Point1, '5.2': scans.Scan5Point2, '5.3': scans.Scan5Point3,
          '5.4': scans.Scan5Point4, '5.5': scans.Scan5Point5, 
          '5.6': scans.Scan5Point6, '5.7': scans.Scan5Point7, 
          '2016b': scans.Scan2016b, 
          '2017a': scans.Scan2017a, '2017b': scans.Scan2017b, 
          '2018a': scans.Scan2018a, '2018b': scans.Scan2018b,
          '2019a': scans.Scan2019a, '2019b': scans.Scan2019b,
          '2020': scans.Scan2020, '2021': scans.Scan2021}

def read_scan(pathnames, dtype=np.int16, join_contiguous=False):
    """ Reads a ScanImage scan.

    Args:
        pathnames: String or list of strings. Pathname(s) or pathname pattern(s) to read.
        dtype: Data-type. Data type of the output array.
        join_contiguous: Boolean. For multiROI scans (2016b and beyond) it will join
            contiguous scanfields in the same depth. No effect in non-multiROI scans. See
            help of ScanMultiROI._join_contiguous_fields for details.

    Returns:
        A Scan object (subclass of BaseScan) with metadata and data. See Readme for details.
    """
    # Expand wildcards
    filenames = expand_wildcard(pathnames)
    if len(filenames) == 0:
        error_msg = 'Pathname(s) {} do not match any files in disk.'.format(pathnames)
        raise PathnameError(error_msg)

    # Read version from one of the tiff files
    with TiffFile(filenames[0]) as tiff_file:
        file_info = tiff_file.pages[0].description + '\n' + tiff_file.pages[0].software
    version = get_scanimage_version(file_info)

    # Select the appropriate scan object
    
    if (version in ['2016b', '2017a', '2017b', '2018a', '2018b', '2019a', '2019b', '2020', '2021'] and
            is_scan_multiROI(file_info)):
        scan = scans.ScanMultiROI(join_contiguous=join_contiguous)
    elif version in _scans:
        scan = _scans[version]()
    else:
        error_msg = 'Sorry, ScanImage version {} is not supported'.format(version)
        raise ScanImageVersionError(error_msg)

    # Read metadata and data (lazy operation)
    scan.read_data(filenames, dtype=dtype)

    return scan

def expand_wildcard(wildcard):
    """ Expands a list of pathname patterns to form a sorted list of absolute filenames.

    Args:
        wildcard: String or list of strings. Pathname pattern(s) to be extended with glob.

    Returns:
        A list of string. Absolute filenames.
    """
    if isinstance(wildcard, str):
        wildcard_list = [wildcard]
    elif isinstance(wildcard, (tuple, list)):
        wildcard_list = wildcard
    else:
        error_msg = 'Expected string or list of strings, received {}'.format(wildcard)
        raise TypeError(error_msg)

    # Expand wildcards
    rel_filenames = [glob(wildcard) for wildcard in wildcard_list]
    rel_filenames = [item for sublist in rel_filenames for item in sublist] # flatten list

    # Make absolute filenames
    abs_filenames = [path.abspath(filename) for filename in rel_filenames]

    # Sort
    sorted_filenames = sorted(abs_filenames, key=path.basename)

    return sorted_filenames

def get_scanimage_version(info):
    """ Looks for the ScanImage version in the tiff file headers.

    Args:
        info: A string. All headers from tiff tags.

    Returns:
        A string. ScanImage version
    """
    pattern = re.compile(r"SI.?\.VERSION_MAJOR = '?(?P<version>[^\s']*)'?")
    match = re.search(pattern, info)
    if match:
        version = match.group('version')
    else:
        raise ScanImageVersionError('Could not find ScanImage version in the tiff header')

    return version

def is_scan_multiROI(info):
    """Looks whether the scan is multiROI in the tiff file headers.

    Args:
        info: A string. All headers from tiff tags.

    Returns:
        A bool. Whether scan is multiroi or not.
    """
    match = re.search(r'hRoiManager\.mroiEnable = (?P<is_multiROI>.)', info)
    is_multiROI = (match.group('is_multiROI') == '1') if match else None
    return is_multiROI



"""Utility functions to check that the key and indices send to __getitem__ are valid."""
import numpy as np

def fill_key(key, num_dimensions):
    """ Fill key with slice(None) (':') until num_dimensions size.

    Args:
        key: tuple of indices or single index. key as received by __getitem__().
        num_dimensions: integer. Total number of dimensions needed.

    Raises:
        IndexError: Too many indices in key: len(key) > num_dimensions.
    """
    # Deal with single valued keys, e.g., scan[:] or scan[0]
    if not isinstance(key, tuple):
        key = (key,)

    # Check key is not larger than num_dimensions
    if len(key) > num_dimensions:
        raise IndexError('too many indices for scan: {}'.format(len(key)))

    # Add missing dimensions
    missing_dimensions = num_dimensions - len(key)
    full_key = tuple(list(key) + [slice(None)] * missing_dimensions)

    return full_key


def check_index_type(axis, index):
    """ Checks that index is an integer, slice or array/list/tuple of integers.

    Args:
        axis: An integer. Axis of the index.
        index: A single index (integer, slice or list/tuple/array of integers).

    Raises:
        TypeError: If index is not integer, slice, or array/list/tuple of integers.
    """
    if not _index_has_valid_type(index): # raise error
        error_msg = ('index {} in axis {} is not an integer, slice or array/list/tuple '
                     'of integers'.format(index, axis))
        raise TypeError(error_msg)


def _index_has_valid_type(index):
    if np.issubdtype(type(index), np.signedinteger): # integer
        return True
    if isinstance(index, slice): # slice
        return True
    if (isinstance(index, (list, tuple)) and
        all(np.issubdtype(type(x), np.signedinteger) for x in index)):  # list or tuple
        return True
    if (isinstance(index, np.ndarray) and np.issubdtype(index.dtype, np.signedinteger)
        and index.ndim == 1):  # array
        return True

    return False


def check_index_is_in_bounds(axis, index, dim_size):
    """ Check that an index is in bounds for the given dimension size. By python indexing
    rules, anything from -dim_size to dim_size-1 is valid.

    Args:
        axis: An integer. Axis of the index.
        index: A single index (integer, slice or list/tuple/array of integers).
        dim_size: Size of the dimension against which the index will be checked.

    Raises:
        TypeError: If index is not either integer, slice, or array.
        IndexError: If index is out of bounds for the specified axis.
    """
    if not _is_index_in_bounds(index, dim_size):
        error_msg = ('index {} is out of bounds for axis {} with size '
                     '{}'.format(index, axis, dim_size))
        raise IndexError(error_msg)


def _is_index_in_bounds(index, dim_size):
    if np.issubdtype(type(index), np.signedinteger):
        return (index in range(-dim_size, dim_size))
    elif isinstance(index, (list, tuple, np.ndarray)):
        return all(x in range(-dim_size, dim_size) for x in index)
    elif isinstance(index, slice):
        return True  # slices never go out of bounds, they are just cropped
    else:
        error_msg = ('index {} is not either integer, slice or array/list/tuple of '
                     'integers'.format(index))
        raise TypeError(error_msg)


def listify_index(index, dim_size):
    """ Generates the list representation of an index for the given dim_size.

    Args:
        index: A single index (integer, slice or list/tuple/array of integers).
        dim_size: Size of the dimension corresponding to the index.

    Returns:
        A list of positive integers. List of indices.

    Raises:
        TypeError: If index is not either integer, slice, or array.
    """
    if np.issubdtype(type(index), np.signedinteger):
        index_as_list = [index] if index >= 0 else [dim_size + index]
    elif isinstance(index, (list, tuple, np.ndarray)):
        index_as_list = [x if x >= 0 else (dim_size + x) for x in index]
    elif isinstance(index, slice):
        start, stop, step = index.indices(dim_size)  # transforms Nones and negative ints to valid slice
        index_as_list = list(range(start, stop, step))
    else:
        error_msg = ('index {} is not integer, slice or array/list/tuple of '
                     'integers'.format(index))
        raise TypeError(error_msg)

    return index_as_list


import argparse
from pathlib import Path

import numpy as np
import fastplotlib as fpl
from mbo_utilities.scanreader.exceptions import PathnameError, FieldDimensionMismatch
from mbo_utilities.scanreader.core import scans, expand_wildcard


def read_scan(pathnames, dtype=np.int16, join_contiguous=False):
    """ Reads a ScanImage scan. """
    # Expand wildcards
    filenames = expand_wildcard(pathnames)
    if len(filenames) == 0:
        error_msg = 'Pathname(s) {} do not match any files in disk.'.format(pathnames)
        raise PathnameError(error_msg)

    scan = ScanMultiROIReordered(join_contiguous=join_contiguous)

    # Read metadata and data (lazy operation)
    scan.read_data(filenames, dtype=dtype)

    return scan


class ScanMultiROIReordered(scans.ScanMultiROI):
    """
    A subclass of ScanMultiROI that ignores the num_fields dimension
    and reorders the output to [time, z, x, y].
    """

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            key = (key,)

        # Call the parent class's __getitem__ with the reordered key
        item = super().__getitem__((0, key[2], key[3], key[1], key[0]))
        if item.ndim == 2:
            return item
        elif item.ndim == 3:
            return np.transpose(item, (2, 0, 1))
        else:
            raise FieldDimensionMismatch('ScanMultiROIReordered.__getitem__')

    @property
    def shape(self):
        return self.num_frames, self.num_channels, self.field_heights[0], self.field_widths[0]

    @property
    def ndim(self):
        return 4


def add_args(parser: argparse.ArgumentParser):
    """
    Add command-line arguments to the parser, dynamically adding arguments
    for each key in the `ops` dictionary.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The argument parser to which arguments are added.

    Returns
    -------
    argparse.ArgumentParser
        The parser with added arguments.
    """
    parser.add_argument('--path', type=str,
                        help='Path to a directory containing raw scanimage tiff files for a single session.')
    parser.add_argument('--version', action='store_true', help='Print the version of the package.')
    return parser


def main():
    parser = argparse.ArgumentParser(description="Preview a scanimage imaging session.")
    parser = add_args(parser)
    args = parser.parse_args()

    # Handle version
    if args.version:
        import mbo_utilities as mbo
        print("lbm_caiman_python v{}".format(mbo.__version__))
        return

    files = [str(f) for f in Path(args.path).glob('*.tif*')]
    scan = read_scan(files, join_contiguous=True)
    iw = fpl.ImageWidget(scan, histogram_widget=False)
    iw.show()
    if fpl.__version__ == "0.2.0":
        fpl.run()
    elif fpl.__version__ == "0.3.0":
        fpl.loop.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import setuptools
import versioneer
from pathlib import Path

install_deps = [
    "tifffile",
    "numpy",
    "scikit-image",
    "fastplotlib[notebook]",
    # "scanreader @ git+https://github.com/atlab/scanreader.git@master#egg=scanreader",
    "PyWavelets",
    "zarr",
]

with open(Path(__file__).parent / "README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mbo_utilities",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Various utilities for the Miller Brain Observatory",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Flynn OConnell",
    author_email="FlynnOConnell@gmail.com",
    license="",
    url="https://github.com/millerbrainobservatory/mbo_utilities",
    keywords="Conda Microscopy ScanImage multiROI Tiff Slurm",
    install_requires=install_deps,
    packages=setuptools.find_packages(exclude=["data", "data.*"]),
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3 :: Only",
    ],
    entry_points={
        "console_scripts": [
            "mbo = mbo_utilities.__main__:main",
        ]
    },
)

